/****************************************************************************
*
*            COPYRIGHT 1990,91,92 BY GRACILIS INC.
*
*			 623 Palace St.
*			 Aurora, Il. 60506
*
*			 (708)-801-8800		Office
*			 (708)-844-0183		(FAX - Support BBS)
*
* GRACILIS, INC., MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS 
* SOFTWARE FOR ANY PURPOSE.  
*
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
* 
* Permission is granted for non-commercial use/distribution only, as long as
* this copyright header is included intact and unaltered.
*
******************************************************************************/


/*
PackeTen PC I/O control port bit definitions:

Address		Value		Result

238			 0x00*		68302 A reset ACTIVE
238			 0x08		68302 A reset INACTIVE

238			 0x01*		68302 B reset ACTIVE
238			 0x09		68302 B reset INACTIVE

238			 0x02*		68302 A Interrupt Forced ACTIVE
238			 0x0A		68302 A Interrupt NORMAL

238			 0x03*		68302 B Interrupt Forced ACTIVE
238			 0x0B		68302 A Interrupt NORMAL

238			 0x04*		IRQ A Clear Forced ACTIVE
238			 0x0C		IRQ A CLear NORMAL

238			 0x05*		IRQ B Clear Forced ACTIVE
238			 0x0D		IRQ B CLear NORMAL

238			 0x06*		Unused
238			 0x0E		Unused

238			 0x07*		Unused
238			 0x0F		Unused

*/

#define PCPORT	0x238		/* Default base port address for a P10 */

/* define this for some printed debug info at run time */
#undef IPC_DEBUG   1


#define OFF	0
#define ON	1

/* offsets into the udcb array for each logical interface */

#define IPC_MC		0	/*  udcb[0] is for IPC to Main Card */
#define IPC_DC		1	/*  udcb[1] is for IPC to Daughter Card */
#define IPC_PC		2	/*  udcb[2] is for IPC to PC Card */

/* Supported driver Types for IPC */
#define IPC_LINK	1	/* an Interprocessor link */

#define P10_MAXIPC	3	/* Maximum number of logical IPC links */

/* Dual Port Memory Base Address for the PackeTen */
#define IPC_MEM_BASE 0xD0000000
#define NUMPROCS     3
#define IPC_RESET  0xff
#define IPC_TX_DEPTH_DEFAULT 32

#ifndef P10_REL2
#define IPC_BUF_SIZE  512
#define IPC_BUF_CNT 5
#else
#define IPC_BUF_SIZE  2048
#define IPC_BUF_CNT 5
#endif

#define	MC_IPC_INT	0x11
#define	DC_IPC_INT	0x11


int ipc_raw __ARGS((struct iface *iface,struct mbuf *data));
#ifdef OLD_KA9Q
int ipc_send __ARGS((struct mbuf *bp,struct iface *iface,int32 gateway,int prec,
	int del,int tput,int rel));
#else
int ipc_send __ARGS((struct mbuf *bp,struct iface *iface,int32 gateway,int tos));
#endif
int ipc_output __ARGS((struct iface *iface,char dest[],char source[],int16 type,
	struct mbuf *data));

/* Driver Control Block */
typedef struct ipc_cblk {
	int16	destproc;		/* Destination IPC proc # 0-2 */
	int16	myproc;			/* My IPC processor # 0-2 */
	int16	rxindex;		/* next receive DP buffer # */
	int16	txindex;		/* next xmit DP buffer # */
	struct	ipcbuf *rxarea;		/* my DP receive array of ipcbufs */
	struct	ipcbuf *txarea;		/* my DP transmit array of ipcbufs */
	unsigned char *txcontrol;	/* my DP transmit control area */
	unsigned char *rxcontrol;	/* my DP receive control area */
	struct drvbuf *xmtq;		/* xmit queue of packets */
	struct drvbuf *xmtqtail;	/* xmit queue of packets */
	int16	txcnt;			/* number of packets in xmtq */
	struct drvbuf *freem;		/* queue of packets to be free'd */
	unsigned long	intcnt;		/* transmit interrupt count */
	unsigned long	txpackcnt;	/* transmitted packet count */

	struct	mbuf *rxbufp;		/* address of current receive buf */
	struct mbuf *rxavailq;		/* Queue of available rx buffers */
	int16  rxavailcount;		/* count of available rx buffers in Q */
	int16  rxavailbufs;		/* num of bufs to try to have on hand */


	unsigned long	rxpackcnt;	/* received packet count */
	unsigned int	nobufs;		/* no memory for pre-allocated */
					/* receive buffers */
	unsigned int	txq_depth;	/* tx pack to be allowed on xmtq at */
					/* any one time */ 
	unsigned int	txflush;	/* tx pack flushed for flow control */
	struct iface *iface;	    	/* For NOS need iface's address */
	struct proc *rxavproc;		/* addr of process for ipc_rxavget() */
	struct proc  *freetxproc; 	/* addr of process for ipc_freetx() */
} IPC_DCB; 


/* Actual Rx buffer format */
struct	ipcbuf	{
			char	active;		/* Buffer full? TRUE/FALSE */
			unsigned char	msbcnt;		/* MSB of char count */
			unsigned char	lsbcnt;		/* LSB */
			char	buffer[IPC_BUF_SIZE];	/* data area */
		};



/* Individual Processor Control/Data area */
struct	ipcproc	{
			char control[NUMPROCS-1];
			struct 	ipcbuf proc_buf[NUMPROCS-1][IPC_BUF_CNT];
		};


/* Global overlay of IPC ram */
typedef struct ipc {
			char			active[NUMPROCS];
			struct	ipcproc	   	proc[NUMPROCS];	/* processor ctl blks */
		   } IPC;



/* The pwait/psignal inter-process communication wake-up array */
/* for getting pre-allocated receive buffers */
extern int  ipc_bufpp[];


/* The pwait/psignal inter-process communication wake-up array */
/* for free-ing transmitted messages */
extern int  ipc_freetxpp[];

extern void ipc_rxavget();
extern void ipc_freetx();

/* misc prototypes for functions in p10_dr.c */
extern void int302mc();
extern void int302dc();
extern void clrint302mc();
extern void clrint302dc();
extern void ipc_mapin();
extern void ipc_mapout();
extern bool ipc_rxint();
extern bool ipc_txint();
extern void ipc_tx_control();
extern int	ipc_raw();
extern int  ipcsend();

/* from p10_at.c */
extern int ipc_attach();
extern int ipc_stop();
extern int doipcstat();
extern int32 ipc_ioctl();
extern long int ghtol();

