/****************************************************************************
*
*            COPYRIGHT 1990,91,92 BY GRACILIS INC.
*
*			 623 Palace St.
*			 Aurora, Il. 60506
*
*			 (708)-801-8800		Office
*			 (708)-844-0183		(FAX - Support BBS)
*
* GRACILIS, INC., MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF THIS 
* SOFTWARE FOR ANY PURPOSE.  
*
* THIS SOFTWARE IS PROVIDED "AS IS" WITHOUT EXPRESS OR IMPLIED WARRANTY.
* 
* Permission is granted for non-commercial use/distribution only, as long as
* this copyright header is included intact and unaltered.
*
******************************************************************************/

/************************************************************************
*
*
* File: p10_dr.c
*
* Rev : 1.0
*
* Description:	
*
*	This File contains routines that implement PackeTen Switch's 
*	inter-processor driver with an interface to KA9Q's TCP/IP suite.
*
*
* Routines: 
*
*	ipc_attach	- Performs attach functions for TCP/IP, i.e.
*		    	  attach the ipc driver to TCP/IP.
*
*	ipc_raw		- Applications' routine for transmitting
*		    	  a message via ipc driver.
*
*	ipc_recv	- Applications' routine for receiving a message
*		      	  from the ipc driver.
*
*	ipc_stat	- Applications' routine for "displaying" the
*		    	  ipc driver statistics.
*
*	ipc_ioctl	- Set driver parameters.
*
*	ipc_stop	- Stop I/O
*	
*	ipc_init	- Initialize the ipc hardware.
*
*	ipc_txint	- Handles transmit complete interrupts.
*
*	ipc_rxint	- Handles receive complete interrupts.
*
*	ipc_entpt	- Interrupt entry point
*
*/

#include <stdio.h>
#include "global.h"
#include "pc.h"
#include "config.h"
#include "mbuf.h"
#include "ax25.h"
#include "iface.h"
#include "lapb.h"
#include "proc.h"
#include "trace.h"
#include "netuser.h"
#include "proc.h"
#include "pktdrvr.h"
#include "gracilis.h"
#include "p10.h"
#include "commands.h"

#include <dos.h>

#ifdef OLD_KA9Q
#else
#define	tprintf printf
#endif

/* interrupt handler */
extern INTERRUPT P10ipc_vec();

#ifdef PACKETEN_DEBUG
long	recvcalls = 0;
long	rawcalls = 0;

#endif

/* Master device control blocks for the Tsync drivers... */
UNI_DCB P10_udcb[P10_MAXIPC] = {
	FALSE, 0, NULL, NULL, NULL, NULL, NULL,	
	FALSE, 0, NULL, NULL, NULL, NULL, NULL,	
	FALSE, 0, NULL, NULL, NULL, NULL, NULL,	
};

INTERRUPT (*P10oldvec) __ARGS((void)); /* Old vector contents */
int16   P10_vec;					/* Vector number */

int16	P10Proc_id = 2;				/* The PC is processor 2!!! */

IPC 	*P10Ipc = (IPC *)IPC_MEM_BASE;
int16	P10Pcport = PCPORT;			/* PackeTen control port address */


INTERRUPT (*P10handler) __ARGS((void)) = { P10ipc_vec };  


/* The pwait/psignal inter-process communication wake-up array */
/* for getting pre-allocated receive buffers */
int  ipc_bufpp[NUMPROCS];


/* The pwait/psignal inter-process communication wake-up array */
/* for free-ing transmitted messages */
int  ipc_freetxpp[NUMPROCS];




/* Send an IP datagram on IPC */
int
#ifdef OLD_KA9Q
ipc_send(bp,iface,gateway,prec,del,tput,rel)
struct mbuf *bp;	/* Buffer to send */
struct iface *iface;	/* Pointer to interface control block */
int32 gateway;		/* IP address of next hop */
int prec;
int del;
int tput;
int rel;
#else
ipc_send(bp,iface,gateway,tos)
struct mbuf *bp;	/* Buffer to send */
struct iface *iface;	/* Pointer to interface control block */
int32 gateway;		/* IP address of next hop */
int tos;
#endif
{
	char *gate = NULL;

	return (*iface->output)(iface,gate,iface->hwaddr,0,bp);
}

/* Send a raw packet INDIRECT!!! This seems silly, */
/* but it is structurally correct as per other NOS devices... */
int
ipc_output(iface,dest,source,type,bp)
struct iface *iface;	/* Pointer to interface control block */
char *dest;		/* Destination Ethernet address */
char *source;		/* Source Ethernet address */
int16 type;		/* Type field */
struct mbuf *bp;	/* mbuf ptr field */
{
	return ((*iface->raw)(iface,bp));
}


/* Send raw packet */
int
ipc_raw(intfp, mbufp)
struct iface *intfp;
struct mbuf *mbufp;
{
	IPC_DCB *sp;
	struct drvbuf *xbufp;
	int16  msgsize;
	char	i_state;
	int16 len_p();

#ifdef PACKETEN_DEBUG
	int16  pullsize;

	rawcalls++;
#endif

#ifdef OLD_KA9Q
	dump(intfp, IF_TRACE_OUT, intfp->type, mbufp);
#else
	dump(intfp,IF_TRACE_OUT,mbufp);
#endif

	sp = (IPC_DCB *)P10_udcb[intfp->dev].dcbp;

	/* check to see if the destination processor is active */
	if ( P10Ipc->active[sp->destproc] == FALSE )
	{
		free_p(mbufp);
		return(-1);
	}


	/* get packet's transmission size */
	msgsize = len_p(mbufp);

	xbufp = (struct drvbuf *)malloc(sizeof(struct drvbuf) + msgsize);

	/* If no memory for buffering... then dump the packet on floor */
	/* and return error to caller */
	if(xbufp == (struct drvbuf *)0)
	{
		free_p(mbufp);
		return(-1);
	}


	xbufp->msgsize = msgsize;
	xbufp->next = (struct drvbuf *)0;

#ifdef PACKETEN_DEBUG
	if ( (pullsize = dqdata(mbufp, &(xbufp->buf), msgsize+1)) != msgsize )
	{   /* somethings wrong.  should have gotten all of message */
		tprintf("ipc_raw error wanted %d bytes pullup-up got only %d\n",
		       msgsize, pullsize);
	}
#else
	/* there will be no check for mismatching msgsize and pullsize */
	dqdata(mbufp, &(xbufp->buf), msgsize);
#endif

	/* now enqueue the message !!! */
	i_state = dirps();
	if ( sp->xmtq == (struct drvbuf *)0 )
	{
		sp->xmtq = sp->xmtqtail = xbufp;
	}
	else
	{
		sp->xmtqtail->next = xbufp;
		sp->xmtqtail = xbufp;
	}
	
	/* bump number of tx packets to in queue */
	sp->txcnt++;
	restore(i_state);


	/* Note ipc_txint must be able to run from ISR level as well */
	/* as from "task" level, since ipc_raw is called from task */
	/* level */
	/* Just call the transmitter: it disables ints when it runs and */
	/* it decides if there is an IPC buffer empty into which to  */
	/* transfer the message */

	if(ipc_txint(sp) != TRUE)
	{
		/* flush a packet check - i.e. cannot let driver suck-up */
		/* alot of memory if packets cannot be xmitted */
		i_state = dirps();
		if( sp->txcnt >= sp->txq_depth )
		{
			xbufp = sp->xmtq;
			sp->xmtq = sp->xmtq->next;
			sp->txcnt--;
			restore(i_state);

			free((char *)xbufp);

			/* make a note of a packet being flushed!! */
			sp->txflush++;
#ifdef PACKETEN_DEBUG
			tprintf("IPC_RAW: No Tx buffers available!!!\n");
			fflush(stdout);
#endif
		}
		else	restore(i_state);
	}

	/* Always Kick the other processor to get him to examin his buffers */
	/* If the destination is the "Main Card"... */
	if(sp->destproc == 0) int302mc();

	else  /* interrupt the daughter */
		int302dc();

	return(0);
}

/********************************************************************
*                       IPC_ENTPT	                           *
*                                                                  *
* File: ipc_driver.c 				                   * 
* Rev : 1.0   						           *
*                                                                  *
********************************************************************
*
*
* DESCRIPTION: 
* -----------
* IPC driver interrupt entry  point. 
* Specific service routines are called from here to handle all IPC
* interrupts
*
* ARGUMENTS:
* ---------
*
* RETURN VALUES:
* -------------
*
* GLOBAL INPUTS:
* -------------
*		
* GLOBAL OUTPUTS:
* --------------
*
* FUNCTIONS CALLED:
* ----------------
*
*
* IO:
* --
*
*
*****************************************************************/


void
ipc_entpt()		/* Interrupt entry pt */
{
IPC_DCB *sp;
int16	dev;
bool	rx_flag,tx_flag;

	/* 	 First clear pending ints at this point...        */
	/* NOTE! This code must be FIRST in interrupt routine, to */
	/*       avoid losing frames which might come in during   */
	/*       processing of other frames.                      */
	/*	 This could result in the generation of Extra     */
	/* 	 interrupts which we have previously handled the  */
	/*	 the data for, so we must handle this case.       */
	/*	 Note also that we don't have to clear the 68302  */
	/*	 int's, they are cleared automatically at INTACK  */
	/*	 time.  We just blindly reset any pending IPC     */
	/*	 ints, since we are going to check for them all   */
	/*	 every time anyway.				  */

    clrint302mc();	/* Main Card */
    clrint302dc();	/* Daughter Card */

    /* determine the interrupting device/devices... */
    for(dev=0;dev<P10_MAXIPC;dev++)
    {
      	if(((dev) != P10Proc_id) && ((sp = (IPC_DCB *)P10_udcb[dev].dcbp) != (IPC_DCB *)0) )
      	{

		rx_flag = ipc_rxint(sp);	/* service possible rx ints */

		tx_flag = ipc_txint(sp);	/* service possible tx ints */

		if((rx_flag == TRUE) || (tx_flag == TRUE))
		{
			/* If the destination is the "Main card"... */
			if(sp->destproc == 0) 
				int302mc();
			else  /* interrupt the daughter */
				int302dc();

	    		/* bump interrupt count */
   			sp->intcnt++;
		}


	}	/* End of "if(sp != (IPC_DCB *) NULLBUF)" */

    }

}

/*=================================================================*/
/* End of ipc_entpt	*/
/*=================================================================*/



/*
 * Receive ISR. 
 * The first receive buffers are pre-allocated
 * in the init routine.  Thereafter, they are filled here, queued out, and a
 * new one acquired.
 */

bool
ipc_rxint(sp)
IPC_DCB *sp;
{
register int16	i;
struct mbuf	*tmpbuf;
unsigned	cnt;
struct	ipcbuf	*rxarea;
int16	rxindex;
bool	rx_flag = FALSE;
char	huge *dest;
char	huge *src;

    /* ptr to this processors IPC control struct */
    rxarea = sp->rxarea;

    /* handle control messages... currently only IPC_RESET */
    if ( *sp->rxcontrol == IPC_RESET )
    {
	P10Ipc->active[P10Proc_id] = TRUE;

	sp->txindex = sp->rxindex = 0;

	*sp->rxcontrol = 0;
	for(i=0;i<IPC_BUF_CNT;i++)
	{
	  /* mark the buffers inactive */
	  rxarea[i].active = (unsigned char)FALSE;
	  rxarea[i].msbcnt = (unsigned char)0;
	  rxarea[i].lsbcnt = (unsigned char)0;
	}
    }


					/* all ints are on frame boundarys */
   /* look thru rx bufs */
   while(rxarea[sp->rxindex].active == TRUE)
   {

    	rxindex = sp->rxindex;

	/* size of rcv'd frame */
	cnt = (unsigned char)rxarea[rxindex].msbcnt;
	cnt = cnt << 8;
	cnt = cnt + (unsigned char)rxarea[rxindex].lsbcnt;

	sp->rxpackcnt++;			/* count this packet */

		
	/* try to get a buffer */ 
	/* from the preallocated rx buffer pool */
	tmpbuf = f_dequeavail(&sp->rxavailq,&sp->rxavailcount);
	if ( tmpbuf  != NULLBUF ) 
	{
	     if ( cnt <= IPC_BUF_SIZE ) 
	     {

		/* read in the data frame */
		dest = (char huge *)(sp->rxbufp->data);
		src = (char huge *)rxarea[rxindex].buffer;

/*
		memcpy((sp->rxbufp->data),rxarea[rxindex].buffer,cnt);
*/
		movedata(FP_SEG(src),FP_OFF(src),FP_SEG(dest),FP_OFF(dest),cnt);

		/* update the byte count */
		sp->rxbufp->cnt += cnt;

#ifdef OLD_KA9Q
		/* point back to original data area */
		sp->rxbufp->data -= sizeof(struct phdr);
#endif
	     }

/*	     else err_mem_prt(sp,cnt);
*/

		enqueue(&Hopper, sp->rxbufp);

		sp->rxbufp = tmpbuf;
	}

	else
	{
		/* We were unable to get a new buffer, so
		   we dump this frame and reuse it's buffer
		*/
		sp->nobufs++;
	}

	/* signal the process which */
	/* allocates the rx buffers */
	psignal(&ipc_bufpp[sp->destproc],1);


/*
		rxarea[i].msbcnt = 0;
		rxarea[i].lsbcnt = 0;
*/
	rxarea[rxindex].active = FALSE;

	if( ++sp->rxindex >= IPC_BUF_CNT) sp->rxindex = 0;

	/* Now notify the other processor we received a frame */	
	rx_flag = TRUE;

   }	/* end of while */

    return(rx_flag);

} /* end of ipc_rxint */


/* Transmit interrupt service routine
 */
bool
ipc_txint(sp)	
IPC_DCB  *sp;
{
	char i_state;
	struct drvbuf *btemp;
	struct	ipcbuf *txarea;
	int16	txindex;
	bool	tx_flag = FALSE;
	char	*xbufp;
	int16	xbytes;

	i_state = dirps();

	txarea = sp->txarea;

	while( (sp->xmtq != (struct drvbuf *)0) && 
	       (txarea[sp->txindex].active == FALSE))
	{

		txindex = sp->txindex;

		/* If a buffer to send, we drop thru here */

		/* set up for next msg to send */
		xbufp = &(sp->xmtq->buf);
		xbytes = sp->xmtq->msgsize;
		
	if(xbytes <= IPC_BUF_SIZE)
	{
		memcpy(txarea[txindex].buffer,xbufp,xbytes);

		txarea[txindex].msbcnt = (unsigned char)(xbytes >> 8);
		txarea[txindex].lsbcnt = (unsigned char)(xbytes & 0x00ff);
		txarea[txindex].active = TRUE;

		tx_flag = TRUE;
	}
		/* put the completed frame on the "to be free'd Q" */

		/* DGL added 4/23/90... */
		sp->txcnt--;

		if ( sp->freem == (struct drvbuf *)0 )
		{
			sp->freem = sp->xmtq;
			sp->xmtq = sp->xmtq->next;
			sp->freem->next = (struct drvbuf *)0;
		}
		else
		{
			/* addr of nxt 1 to free*/
			btemp = sp->xmtq;

			/* nxt one to be xmitted... */
			sp->xmtq = sp->xmtq->next;

			/* queue the one to be free'd */
			btemp->next = sp->freem;
			sp->freem = btemp;
		}

		if ( sp->xmtq == (struct drvbuf *)0 ) 
			sp->xmtqtail = (struct drvbuf *)0;

		/* new for NOS */
		/* signal the process which is supposed to */
		/* free the transmitted message */
		psignal(&ipc_freetxpp[sp->destproc],1);

		sp->txpackcnt++;
		
		if(++sp->txindex >= IPC_BUF_CNT) sp->txindex = 0;

	} /* end while */
	restore(i_state);
	return(tx_flag);
}


/*
 *	generate an interrupt to 68302 mc
 */
void
int302mc()
{
char	i_state;

	i_state = dirps();

	outportb(P10Pcport, 0x02);	/* force 68302 A interrupt ACTIVE */
	outportb(P10Pcport, 0x0a);	/* release 68302 A interrupt */

	restore(i_state);
}

/*
 *	generate an interrupt to 68302  dc
 */
void
int302dc()
{
char	i_state;

	i_state = dirps();

	outportb(P10Pcport, 0x03);	/* force 68302 A interrupt ACTIVE */
	outportb(P10Pcport, 0x0b);	/* release 68302 A interrupt */

	restore(i_state);
}

/*
 *	clear 68302 mc interrupt
 *	this must be done within the PC interrupt routine
 */
void
clrint302mc()
{
char	i_state;

	i_state = dirps();

	outportb(P10Pcport, 0x04);	/* IRQ A Clear Forced ACTIVE */
	outportb(P10Pcport, 0x0c);	/* IRQ A CLear NORMAL */

	restore(i_state);
}

/*
 *	clear 68302 dc interrupt
 *	this must be done within the PC interrupt routine
 */
void
clrint302dc()
{

char	i_state;

	i_state = dirps();

	outportb(P10Pcport, 0x05);	/* IRQ B Clear Forced ACTIVE */
	outportb(P10Pcport, 0x0d);	/* IRQ B CLear NORMAL */

	restore(i_state);
}

/*
 *	enable shared memory for a PackeTen as specified by
 *	the I/O port address for it's control port.
 *	this must be done in order to access the IPC memory
 */
void
ipc_mapin(port)
int16	port;
{
char	i_state;

	i_state = dirps();

	outportb(port, 0x0f);	/* enable dual port memory */

	restore(i_state);
} 

/*
 *	disable shared memory for a PackeTen as specified by
 *	the I/O port address for it's control port.
 */
void
ipc_mapout(port)
int16	port;
{
char	i_state;

	i_state = dirps();

	outportb(port, 0x07);	/* disable dual port memory */

	restore(i_state);
} 




/*********************************************************
 * GRACILIS - These are separate processes for buffer mgmt
 * 	          for the PackeTen IPC drivers.
 *********************************************************/

/* Pre-allocate receive buffers for the IPC receive handler  */
void
ipc_rxavget(dev, p1, p2)			
int	dev;			/* device number */
void	*p1;			/* not used */
void	*p2;			/* not used */
{
char	i_state;
IPC_DCB *dcbp;
int16 cur_rxavail, min_rxavail, destproc;


	/* Grace need better error handling? */
	/* killself if device is not attached yet */
	if  ((dev < 0) || (dev >= P10_MAXIPC )) 		
		return;
	else if (P10_udcb[dev].attached == FALSE)
		return;

	dcbp = (IPC_DCB *)P10_udcb[dev].dcbp;
	destproc = dcbp->destproc;

	for (;;)
	{

		/* wait for the driver to signal it needs */
		/* a look at its pre-allocated receive buffers */
		pwait(&ipc_bufpp[destproc]);


		i_state = dirps();

		cur_rxavail = dcbp->rxavailcount;
		min_rxavail = dcbp->rxavailbufs;

		restore(i_state);

	
		/* see if more receive buffers need allocating */
		if ( cur_rxavail < min_rxavail )
		{
			f_getavail(&(dcbp->rxavailq),
			           (int16)(min_rxavail - cur_rxavail),
			           IPC_BUF_SIZE, 
				   &(dcbp->rxavailcount), dcbp->iface,
				   (bool)FALSE );
		}
	} /* end of for(;;) loop */

} /* end of ipc_rxavget() */





/* This process is used to return transmitted */
/* buffers to the free memory pool */
void
ipc_freetx(dev, p1, p2)			
int	dev;			/* device number */
void	*p1;			/* not used */
void	*p2;			/* not used */
{
char	i_state;
int16	destproc;
IPC_DCB *dcbp;
struct drvbuf *xbufp, *btemp;


	/* Grace need better error handling? */
	/* killself if device is not attached yet */
	if  ((dev < 0) || (dev >= P10_MAXIPC)) 		
		return;
	else if (P10_udcb[dev].attached == FALSE)
		return;

	/* driver control block */
	dcbp = (IPC_DCB *)P10_udcb[dev].dcbp;
	destproc = dcbp->destproc;

	for (;;)
	{

		/* wait for driver to signal that there are transmitted */
		/* messages to be freed */
		pwait(&ipc_freetxpp[destproc]);


		/* Do not do memory free-ups while interrupts are */
		/* disabled! Here xbufp is a "to be free'd"  */
		/* queue pointer" */

		i_state = dirps();
		xbufp = dcbp->freem;
		dcbp->freem = (struct drvbuf *)0;
		restore(i_state);

		/* now free memory of all the xmitted messages */
		while ( xbufp != (struct drvbuf *)0 )
		{
			btemp = xbufp->next;

			free((char *)xbufp);


			xbufp = btemp;
		}
	} /* end of for(;;) loop */

} /* end of ipc_freetx() */


